/*===============================================================================*
 * Copyright (C) 2020                                                            *
 *                                                                               *
 * Authors: Dario Menasce                        *
 *                                                                               *
 * INFN: Piazza della Scienza 3, Edificio U2, Milano, Italy 20126                *
 *                                                                               *
 * This program is free software: you can redistribute it and/or modify          *
 * it under the terms of the GNU General Public License as published by          *
 * the Free Software Foundation, either version 3 of the License, or             *
 * (at your option) any later version.                                           *
 *                                                                               *
 * This program is distributed in the hope that it will be useful,               *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
 * GNU General Public License for more details.                                  *
 *                                                                               *
 * You should have received a copy of the GNU General Public License             *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.         *
 ================================================================================*/

#include <analisi.h>

Double_t logistic(Double_t *x, Double_t *par)
{
   Float_t xx =x[0];
   Double_t f = par[0]/(1+exp(-par[1]*(xx-par[2])));
   return f;
}
//==================================================
Analisi::Analisi(string TFileName)
{
 nBins_     = 50        ;
 xmi_       = 0         ;
 xma_       = nBins_    ;
 TFileName_ = TFileName ;
 addAll_    = false     ;
 binMap_                ; 
   
 listaTipoDato_  ["data_"                       ] = 1 ;
 listaTipoDato_  ["stato_"                      ] = 1 ;
 listaTipoDato_  ["codice_regione_"             ] = 1 ;
 listaTipoDato_  ["denominazione_regione_"      ] = 1 ;    
 listaTipoDato_  ["latitude_"                   ] = 1 ;                 
 listaTipoDato_  ["longitude_"                  ] = 1 ;                   
 listaTipoDato_  ["ricoverati_con_sintomi_"     ] = 1 ;      
 listaTipoDato_  ["terapia_intensiva_"          ] = 1 ;        
 listaTipoDato_  ["totale_ospedalizzati_"       ] = 1 ;      
 listaTipoDato_  ["isolamento_domiciliare_"     ] = 1 ;    
 listaTipoDato_  ["totale_attualmente_positivi_"] = 1 ;
 listaTipoDato_  ["nuovi_attualmente_positivi_" ] = 1 ;
 listaTipoDato_  ["dimessi_guariti_"            ] = 1 ;          
 listaTipoDato_  ["deceduti_"                   ] = 1 ;                
 listaTipoDato_  ["totale_casi_"                ] = 1 ;              
 listaTipoDato_  ["tamponi_"                    ] = 1 ;                   

 listaEstesaTipi_["data_"                       ] = "Data"                                     ;
 listaEstesaTipi_["stato_"                      ] = "Stato"                                    ;
 listaEstesaTipi_["codice_regione_"             ] = "Codice Regione"                           ;
 listaEstesaTipi_["denominazione_regione_"      ] = "Regione"                                  ;
 listaEstesaTipi_["latitude_"                   ] = "Latitudine"                               ;
 listaEstesaTipi_["longitude_"                  ] = "Longitudine"                              ;
 listaEstesaTipi_["ricoverati_con_sintomi_"     ] = "Ricoverati con sintomi"                   ;
 listaEstesaTipi_["terapia_intensiva_"          ] = "Ricoverati in terapia intensiva"          ; 
 listaEstesaTipi_["totale_ospedalizzati_"       ] = "Totale ospedalizzati"                     ;
 listaEstesaTipi_["isolamento_domiciliare_"     ] = "In isolamento domiciliare"                ;
 listaEstesaTipi_["totale_attualmente_positivi_"] = "Totale positivi"                          ;
 listaEstesaTipi_["nuovi_attualmente_positivi_" ] = "Totale nuovi positivi alla data corrente" ; 
 listaEstesaTipi_["dimessi_guariti_"            ] = "Guariti o dimessi"                        ;
 listaEstesaTipi_["deceduti_"                   ] = "Deceduti"                                 ;
 listaEstesaTipi_["totale_casi_"                ] = "Totale dei casi"                          ;
 listaEstesaTipi_["tamponi_"                    ] = "Numero tamponi effettuati"                ;      
  
 cout << __LINE__ << "] Lista delle tipologie considerate" << endl ;
 for(binMapDef_::iterator it =listaTipoDato_.begin();
                          it!=listaTipoDato_.end();
                          it++)
 {
  cout << __LINE__ << "] " << it->first << endl ;
 }    

 cout << __LINE__ << "] "                                         << endl ;
 cout << __LINE__ << "] Lista estesa delle tipologie considerate" << endl ;
 for(strMapDef_::iterator it =listaEstesaTipi_.begin();
                          it!=listaEstesaTipi_.end()  ;
                          it++)
 {
  cout << __LINE__ << "] " << it->second << endl ;
 }    
}
//==================================================
void Analisi::makeBranch(void)
{
 cout << __LINE__ << "] Making branch" << endl ;
 file_    = new TFile(TFileName_.c_str());
 branch_  = (TTree*)file_->Get("branchRegione");
 if( !branch_ )
 {
  cout << __LINE__ << "Branch not found!" << endl ;
 }

 branch_->SetBranchAddress("data"                       ,&data_                       );
 branch_->SetBranchAddress("stato"                      ,&stato_                      );
 branch_->SetBranchAddress("codice_regione"             ,&codice_regione_             );
 branch_->SetBranchAddress("denominazione_regione"      ,&denominazione_regione_      );
 branch_->SetBranchAddress("latitude"                   ,&latitude_                   );
 branch_->SetBranchAddress("longitude"                  ,&longitude_                  );
 branch_->SetBranchAddress("ricoverati_con_sintomi"     ,&ricoverati_con_sintomi_     );
 branch_->SetBranchAddress("terapia_intensiva"          ,&terapia_intensiva_          );
 branch_->SetBranchAddress("totale_ospedalizzati"       ,&totale_ospedalizzati_       );
 branch_->SetBranchAddress("isolamento_domiciliare"     ,&isolamento_domiciliare_     );
 branch_->SetBranchAddress("totale_attualmente_positivi",&totale_attualmente_positivi_);
 branch_->SetBranchAddress("nuovi_attualmente_positivi" ,&nuovi_attualmente_positivi_ );
 branch_->SetBranchAddress("dimessi_guariti"            ,&dimessi_guariti_            );
 branch_->SetBranchAddress("deceduti"                   ,&deceduti_                   );
 branch_->SetBranchAddress("totale_casi"                ,&totale_casi_                );
 branch_->SetBranchAddress("tamponi"                    ,&tamponi_                    );
}
//=======================================================================
void Analisi::readTree(bool dump)
{
 cout << __LINE__ << "] Reading tree..." << endl ;
 nentries_ = branch_->GetEntries();
 cout << __LINE__ << "] Totaledati trovati: " << nentries_ << endl ;

 int bin = 0 ;
 for (Long64_t i=0; i<nentries_; i++) 
 {         
  branch_->GetEntry(i);
  mapData_[*data_][*denominazione_regione_]["ricoverati_con_sintomi_"     ] = ricoverati_con_sintomi_     ;
  mapData_[*data_][*denominazione_regione_]["terapia_intensiva_"          ] = terapia_intensiva_          ;
  mapData_[*data_][*denominazione_regione_]["totale_ospedalizzati_"       ] = totale_ospedalizzati_       ;
  mapData_[*data_][*denominazione_regione_]["isolamento_domiciliare_"     ] = isolamento_domiciliare_     ;
  mapData_[*data_][*denominazione_regione_]["totale_attualmente_positivi_"] = totale_attualmente_positivi_;
  mapData_[*data_][*denominazione_regione_]["nuovi_attualmente_positivi_" ] = nuovi_attualmente_positivi_ ;
  mapData_[*data_][*denominazione_regione_]["dimessi_guariti_"            ] = dimessi_guariti_            ;
  mapData_[*data_][*denominazione_regione_]["deceduti_"                   ] = deceduti_                   ;
  mapData_[*data_][*denominazione_regione_]["totale_casi_"                ] = totale_casi_                ;
  mapData_[*data_][*denominazione_regione_]["tamponi_"                    ] = tamponi_                    ;
  cout << __LINE__ << "] " << *data_ << " Regione: " << *denominazione_regione_ << "\t" << totale_casi_ << endl ;

  listaRegioni_ [*denominazione_regione_] = 1 ;
  if( binMap_.find(*data_) == binMap_.end()) 
  {
   bin++ ;
   binMap_[*data_] = bin ;
  }
 }
 
 cout << __LINE__ << "] "                                  << endl ;
 cout << __LINE__ << "] Lista delle regioni considerate: " << endl ;
 for(binMapDef_::iterator it=listaRegioni_.begin(); it!=listaRegioni_.end(); ++it)
 {
  cout << __LINE__ << "] Regione: " << it->first << endl ;
 }
}
//=======================================================================
void Analisi::dumpAllData(bool dump)
{
 if(!dump) return ;
 for(datMapDef_::iterator it=mapData_.begin(); it!=mapData_.end(); ++it)
 {
  cout << __LINE__ << "] Data: " << it->first << endl ;
  for(ssrMapDef_::iterator jt=it->second.begin(); jt!=it->second.end(); jt++)
  {
   cout << __LINE__ << "]   Regione: " << jt->first << endl ;
   for(binMapDef_::iterator kt=jt->second.begin(); kt!=jt->second.end(); kt++)
   {
    cout << __LINE__ << "]      : " << kt->first << " = " << kt->second << endl ;
   } 
  }
 }
}
//=======================================================================
Analisi::vDef Analisi::getDates( void )
{
 vDef dates ;
 for(datMapDef_::iterator it=mapData_.begin(); it!=mapData_.end(); ++it)
 {
  dates.push_back(it->first) ;
 }
 return dates ;
}
//=======================================================================
void Analisi::bookHistograms(vDef regione, vDef tipologia)
{
 cout << __LINE__ << "] Booking histograms" << endl ;

 if( regione[0] == "all" ) 
 {
  cout << __LINE__ << "] Richiesta somma delle tipologie a livello nazionale" << endl ;
  addAll_ = true ;
  internalTipologia_ = tipologia ;
 }
 if( (regione.size() + tipologia.size()) >= 30) 
 {
  cout << __LINE__ << "] Spiacente, il massimo numero di istogrammi gestibile da questo programma e' 30" << endl  ;
  exit(0) ;
 }

 stringstream sn;
 stringstream st;
 for(binMapDef_::iterator it=listaRegioni_.begin(); it!=listaRegioni_.end(); ++it)
 {
  bool rejectRegione = true ;
  for(int i=0; i<regione.size(); ++i)
  {
   if(it->first == regione[i]) rejectRegione = false ;
  }
  if( regione[0] == "all" ) rejectRegione = false ;
  if( rejectRegione ) {continue;}
  for(binMapDef_::iterator jt=listaTipoDato_.begin(); jt!=listaTipoDato_.end(); ++jt)
  {
   bool rejectTipologia = true ;
   for(int i=0; i<tipologia.size(); ++i)
   {
    if(jt->first == tipologia[i]) rejectTipologia = false ;
   }
   if( rejectTipologia ) {continue;}
   if( jt->first == "data_"                  ||
       jt->first == "stato_"                 ||
       jt->first == "codice_regione_"        ||
       jt->first == "denominazione_regione_" ||
       jt->first == "latitude_"              ||
       jt->first == "longitude_"              ) {continue;}
   sn.str(""); sn << it->first << "_" << jt->first ;
   st.str(""); st << "Regione " << it->first << " - Tipo dati: " << listaEstesaTipi_[jt->first] ;
   hisMap_[it->first][jt->first] = (new TH1F(sn.str().c_str(),st.str().c_str(),nBins_,xmi_,xma_)) ;
   hisMap_[it->first][jt->first]->SetMarkerStyle( 20) ;
   hisMap_[it->first][jt->first]->SetMarkerSize (0.5) ;
   hisMap_[it->first][jt->first]->SetMarkerColor(  2) ;
  }
 }
 
}

//=======================================================================
bool Analisi::wrongField(string tipologia)
{
 if( tipologia == "data_"                  ||
     tipologia == "stato_"                 ||
     tipologia == "codice_regione_"        ||
     tipologia == "denominazione_regione_" ||
     tipologia == "latitude_"              ||
     tipologia == "longitude_"              ) {return true;}
 return false ;
}
//=======================================================================
void Analisi::fillHistograms(vDef regione, vDef tipologia)
{
 cout << __LINE__ << "] Riempimento istogrammi" << endl ;
 for(datMapDef_::iterator it=mapData_.begin(); it!=mapData_.end(); ++it)
 {
  int bin = binMap_[it->first] ;
//  cout << __LINE__ << "] Data: " << it->first << endl ;
  for(ssrMapDef_::iterator jt=it->second.begin(); jt!=it->second.end(); jt++)
  {
   bool rejectRegione = true ;
   for(int i=0; i<regione.size(); ++i)
   {
    if(jt->first == regione[i]) rejectRegione = false ;
   }
   if( regione[0] == "all" ) rejectRegione = false ;
   if( rejectRegione ) {continue;}
//   cout << __LINE__ << "]   Found regione " << jt->first << endl ;
   for(binMapDef_::iterator lt=jt->second.begin(); lt!=jt->second.end(); lt++)
   {
    bool rejectTipologia = true ;
    for(int i=0; i<tipologia.size(); ++i)
    {
     if(lt->first == tipologia[i]) rejectTipologia = false ;
    }
//    cout << __LINE__ << "] regione " << jt->first << " " << lt-first << endl ;
    if( rejectTipologia ) {continue;}
//     cout << __LINE__ << "] it->first: " << it->first 
//                       << " jt->first: " << jt->first 
//                       << " lt->first: " << lt->first 
//                       << " bin: "       << bin 
//                       << " = "          << lt->second << endl ;
//    cout << __LINE__ << "] ptr : " << hisMap_[jt->first][lt->first] << endl ;
//    cout << __LINE__ << "] name: " << hisMap_[jt->first][lt->first]->GetName() << endl ;
    hisMap_[jt->first][lt->first]->SetBinContent(bin,lt->second) ;
   }
  }
 }
}
//=======================================================================
void Analisi::draw(bool doFits, bool optLogy)
{
 int pad   = 1 ;
 int count = 0 ;

 gStyle->SetOptStat(0) ;
 gStyle->SetOptFit (0) ;
 if( optLogy ) gStyle->SetOptLogy(1) ;
 else          gStyle->SetOptLogy(0) ;

 if( addAll_ ) 
 {
  cout << __LINE__ << "] Somma degli istogrammi" << endl ;
  vector<TH1*> hTotale ;
  for(hisMapDef_::iterator it=hisMap_.begin(); it!=hisMap_.end(); it++)
  {
   cout << __LINE__ << "] Regione: " << it->first << endl ;
   for(ptrMapDef_::iterator jt=it->second.begin(); jt!=it->second.end(); jt++)
   {
    stringstream s ;
    s << jt->first << "Nazionale" ;
    hTotale.push_back(new TH1F(s.str().c_str(),s.str().c_str(),nBins_,xmi_,xma_)) ;
    cout << __LINE__ << "] " << hTotale.back()->GetName() << endl ;
    count++ ;
   }
  }
 }
 else
 {
  cout << __LINE__ << "] Istogrammi proiettati sul canvas" << endl ;
  for(hisMapDef_::iterator it=hisMap_.begin(); it!=hisMap_.end(); it++)
  {
   cout << __LINE__ << "] Regione: " << it->first << endl ;
   for(ptrMapDef_::iterator jt=it->second.begin(); jt!=it->second.end(); jt++)
   {
    count++ ;
   }
  }

  int nx = 1 ;
  int ny = 2 ;
  if( count ==  1 ) {nx = 1; ny = 1;} ;
  if( count ==  2 ) {nx = 2; ny = 1;} ;
  if( count ==  3 ) {nx = 3; ny = 1;} ;
  if( count ==  4 ) {nx = 2; ny = 2;} ;
  if( count ==  6 ) {nx = 3; ny = 2;} ;
  if( count ==  7 ) {nx = 3; ny = 3;} ;
  if( count ==  8 ) {nx = 3; ny = 3;} ;
  if( count ==  9 ) {nx = 3; ny = 3;} ;
  if( count == 10 ) {nx = 4; ny = 3;} ;
  if( count == 11 ) {nx = 4; ny = 3;} ;
  if( count == 12 ) {nx = 4; ny = 3;} ;
  if( count == 13 ) {nx = 4; ny = 4;} ;
  if( count == 14 ) {nx = 4; ny = 4;} ;
  if( count == 15 ) {nx = 4; ny = 4;} ;
  if( count == 16 ) {nx = 4; ny = 4;} ;
  if( count == 17 ) {nx = 5; ny = 4;} ;
  if( count == 18 ) {nx = 5; ny = 4;} ;
  if( count == 19 ) {nx = 5; ny = 4;} ;
  if( count == 20 ) {nx = 5; ny = 4;} ;
  if( count == 21 ) {nx = 5; ny = 5;} ;
  if( count == 22 ) {nx = 5; ny = 5;} ;
  if( count == 23 ) {nx = 5; ny = 5;} ;
  if( count == 24 ) {nx = 5; ny = 5;} ;
  if( count == 25 ) {nx = 5; ny = 5;} ;
  if( count == 26 ) {nx = 6; ny = 5;} ;
  if( count == 27 ) {nx = 6; ny = 5;} ;
  if( count == 28 ) {nx = 6; ny = 5;} ;
  if( count == 29 ) {nx = 6; ny = 5;} ;
  if( count >= 30 ) return ;

  stringstream ss ;

  canvas_ = new TCanvas("c","c",1200,1000) ;
  canvas_->Divide(nx,ny) ;
  cout << __LINE__ << "] Numero istogrammi: " << count << " Divide: " << nx << "x" << ny << endl ;
 
  for(hisMapDef_::iterator it=hisMap_.begin(); it!=hisMap_.end(); it++)
  {
   cout << __LINE__ << "] Regione: " << it->first << endl ;
   for(ptrMapDef_::iterator jt=it->second.begin(); jt!=it->second.end(); jt++)
   {
    erfFunc_.push_back(new TF1("erf","[0]+[1]*TMath::Erf((x-[2])/[3])",0,100)) ;
    expFunc_.push_back(new TF1("exp","[0]*TMath::Exp((x-[1])/[2])"    ,0,100)) ;
    erfFunc_.back()->SetLineColor(2) ;
    expFunc_.back()->SetLineColor(4) ;

    erfFunc_.back()->SetParameter(0, 1) ;
    erfFunc_.back()->SetParameter(1, 1) ;
    erfFunc_.back()->SetParameter(2, 1) ;
    erfFunc_.back()->SetParameter(3, 1) ;
    
    expFunc_.back()->SetParameter(0, 1) ;
    expFunc_.back()->SetParameter(1, 1) ;
    expFunc_.back()->SetParameter(2, 1) ;

    canvas_->cd(pad++); 
//jt->second->GetXaxis()->SetRangeUser(0,25) ;
//jt->second->GetYaxis()->SetRangeUser(0,70000) ;
    jt->second->Draw("E1") ;

    if( !doFits ) {continue;}

    TFitResultPtr ERFptr = jt->second->Fit(erfFunc_.back(),"SB","",0,21.6) ;
    TFitResultPtr EXPptr = jt->second->Fit(expFunc_.back(),"SB","",0,21.6) ;
    erfFunc_.back()->Draw("SAME") ;
    expFunc_.back()->Draw("SAME") ;

    TLegend *legERF = new TLegend(0.6, 0.1, 0.9, 0.25);
    legERF->SetHeader("Function: [0]+[1]*TMath::Erf((x-[2])/[3])");
    ss.str(""); ss << "chi2: " << ERFptr->Chi2()/erfFunc_.back()->GetNDF() ;         
    legERF->AddEntry(erfFunc_.back(), ss.str().c_str(), "L");
    ss.str(""); ss << "[0]: "  << ERFptr->Parameter(0) << "+/-" << ERFptr->ParError(0);         
    legERF->AddEntry(erfFunc_.back(), ss.str().c_str(), "L");
    ss.str(""); ss << "[1]: "  << ERFptr->Parameter(1) << "+/-" << ERFptr->ParError(1);         
    legERF->AddEntry(erfFunc_.back(), ss.str().c_str(), "L");
    ss.str(""); ss << "[2]: "  << ERFptr->Parameter(2) << "+/-" << ERFptr->ParError(2);         
    legERF->AddEntry(erfFunc_.back(), ss.str().c_str(), "L");
    ss.str(""); ss << "[3]: "  << ERFptr->Parameter(3) << "+/-" << ERFptr->ParError(3);         
    legERF->AddEntry(erfFunc_.back(), ss.str().c_str(), "L");
    legERF->Draw("SAME");

    cout << __LINE__ << "] ERF chi2 / NDOF: " << ERFptr->Chi2() << endl ;

    TLegend *legEXP = new TLegend(0.6, 0.3, 0.9, 0.45);
    legEXP->SetHeader("Function: [0]*TMath::Exp((x-[1])/[2])");
    ss.str(""); ss << "chi2: " << EXPptr->Chi2()/expFunc_.back()->GetNDF() ;         
    legEXP->AddEntry(expFunc_.back(), ss.str().c_str(), "L");
    ss.str(""); ss << "[0]: "  << EXPptr->Parameter(0) << " +/- " << EXPptr->ParError(0);         
    legEXP->AddEntry(expFunc_.back(), ss.str().c_str(), "L");
    ss.str(""); ss << "[1]: "  << EXPptr->Parameter(1) << " +/- " << EXPptr->ParError(1);         
    legEXP->AddEntry(expFunc_.back(), ss.str().c_str(), "L");
    ss.str(""); ss << "[2]: "  << EXPptr->Parameter(2) << " +/- " << EXPptr->ParError(2);         
    legEXP->AddEntry(expFunc_.back(), ss.str().c_str(), "L");
    legEXP->AddEntry(expFunc_.back(), " ", "P");
    legEXP->Draw("SAME");

    cout << __LINE__ << "] EXP chi2 / NDOF: " << EXPptr->Chi2() << endl ;
   }
  }
  cout << __LINE__ << "] Done!" << endl ;
 }
} 

//=======================================================================

Analisi::paiDef Analisi::fitLogistic(TH1F * plot)
{
 TF1 * logFunc = new TF1("log",logistic,xmi_,xma_,3) ;

 logFunc->SetLineColor(6   ) ;
 logFunc->SetParameter(0, 1) ;
 logFunc->SetParameter(1, 1) ;
 logFunc->SetParameter(2, 1) ;
 TFitResultPtr LOGptr = plot->Fit(logFunc,"SBN");

 Analisi::paiDef plotResults = make_pair(logFunc,LOGptr) ;

 return plotResults ;
}

//=======================================================================
void Analisi::plotAndFit(TH1F * plot, bool doFit)
{
 cout << __LINE__ << "[ plotting and fitting" << endl ;
 
 gStyle->SetOptStat(0) ;
 gStyle->SetOptFit (0) ;
 gStyle->SetOptLogy(0) ;

// plot->GetYaxis()->SetRangeUser(0.,16500.) ;
  
 plot->Draw("E1") ;

 if( !doFit ) {return;}

 TF1 * erfFunc = new TF1("erf","[0]+[1]*TMath::Erf((x-[2])/[3])",xmi_,xma_) ;
 TF1 * expFunc = new TF1("exp","[0]*TMath::Exp((x-[1])/[2])"    ,xmi_,xma_) ;
 TF1 * logFunc = new TF1("log",logistic,xmi_,xma_,3) ;
 cout << __LINE__ << "] ptr: " << logFunc << endl ;
 erfFunc->SetLineColor(2) ;
 expFunc->SetLineColor(4) ;
 logFunc->SetLineColor(6) ;

 erfFunc->SetParameter(0, 1) ;
 erfFunc->SetParameter(1, 10) ;
 erfFunc->SetParameter(2, 1) ;
 erfFunc->SetParameter(3, 1) ;
 
 expFunc->SetParameter(0, 1) ;
 expFunc->SetParameter(1, 1) ;
 expFunc->SetParameter(2, 1) ;

 logFunc->SetParameter(0, 1) ;
 logFunc->SetParameter(1, 1) ;
 logFunc->SetParameter(2, 1) ;


 TFitResultPtr ERFptr = plot->Fit(erfFunc,"SB","",0,21.6) ;
 TFitResultPtr EXPptr = plot->Fit(expFunc,"SB","",0,21.6) ;
 TFitResultPtr LOGptr = plot->Fit(logFunc);
 
 erfFunc->Draw("SAME") ;
 expFunc->Draw("SAME") ;
 logFunc->Draw("SAME") ;


 stringstream ss ;

 TLegend *legEXP = new TLegend(0.6, 0.3, 0.9, 0.45);
 legEXP->SetHeader("Exponential: [0]*TMath::Exp((x-[1])/[2])");
 ss.str(""); ss << "chi2/NDOF: " << EXPptr->Chi2()/expFunc->GetNDF() ;         
 legEXP->AddEntry(expFunc, ss.str().c_str(), "L");
 ss.str(""); ss << "[0]: "  << EXPptr->Parameter(0) << " +/- " << EXPptr->ParError(0);         
 legEXP->AddEntry(expFunc, ss.str().c_str(), "L");
 ss.str(""); ss << "[1]: "  << EXPptr->Parameter(1) << " +/- " << EXPptr->ParError(1);         
 legEXP->AddEntry(expFunc, ss.str().c_str(), "L");
 ss.str(""); ss << "[2]: "  << EXPptr->Parameter(2) << " +/- " << EXPptr->ParError(2);         
 legEXP->AddEntry(expFunc, ss.str().c_str(), "L");
 legEXP->AddEntry(expFunc, " ", "P");
 legEXP->Draw("SAME");

 TLegend *legERF = new TLegend(0.6, 0.45, 0.9, 0.60);
 legERF->SetHeader("ERF: [0]+[1]*TMath::Erf((x-[2])/[3])");
 ss.str(""); ss << "chi2/NDOF: " << ERFptr->Chi2()/erfFunc->GetNDF() ;         
 legERF->AddEntry(erfFunc, ss.str().c_str(), "L");
 ss.str(""); ss << "[0]: "  << ERFptr->Parameter(0) << "+/-" << ERFptr->ParError(0);         
 legERF->AddEntry(erfFunc, ss.str().c_str(), "L");
 ss.str(""); ss << "[1]: "  << ERFptr->Parameter(1) << "+/-" << ERFptr->ParError(1);         
 legERF->AddEntry(erfFunc, ss.str().c_str(), "L");
 ss.str(""); ss << "[2]: "  << ERFptr->Parameter(2) << "+/-" << ERFptr->ParError(2);         
 legERF->AddEntry(erfFunc, ss.str().c_str(), "L");
 ss.str(""); ss << "[3]: "  << ERFptr->Parameter(3) << "+/-" << ERFptr->ParError(3);         
 legERF->AddEntry(erfFunc, ss.str().c_str(), "L");
 legERF->Draw("SAME");

//  TLegend *legLOG = new TLegend(0.6, 0.1, 0.9, 0.25);
//  legLOG->SetHeader("Logistic: par[0]/(1+exp(-par[1]*(xx-par[2]))");
//  ss.str(""); ss << "chi2/NDOF: " << LOGptr->Chi2()/logFunc->GetNDF() ;         
//  legLOG->AddEntry(logFunc, ss.str().c_str(), "L");
//  ss.str(""); ss << "[0]: "  << LOGptr->Parameter(0) << "+/-" << LOGptr->ParError(0);         
//  legLOG->AddEntry(logFunc, ss.str().c_str(), "L");
//  ss.str(""); ss << "[1]: "  << LOGptr->Parameter(1) << "+/-" << LOGptr->ParError(1);         
//  legLOG->AddEntry(logFunc, ss.str().c_str(), "L");
//  ss.str(""); ss << "[2]: "  << LOGptr->Parameter(2) << "+/-" << LOGptr->ParError(2);         
//  legLOG->AddEntry(logFunc, ss.str().c_str(), "L");
//  ss.str(""); ss << "[3]: "  << LOGptr->Parameter(3) << "+/-" << LOGptr->ParError(3);         
//  legLOG->AddEntry(logFunc, ss.str().c_str(), "L");
//  legLOG->Draw("SAME");

 cout << __LINE__ << "] EXP chi2 / NDOF: " << EXPptr->Chi2() << " (NODF: " << expFunc->GetNDF() << ")" << endl ;
 cout << __LINE__ << "] ERF chi2 / NDOF: " << ERFptr->Chi2() << " (NODF: " << erfFunc->GetNDF() << ")" << endl ;
  
// delete erfFunc ;
// delete expFunc ;
}
//=======================================================================
void Analisi::listaRegioni(void)
{
 cout << __LINE__ << "] Lista delle regioni disponibili" << endl ;
 vector<string> regs ;
 for(binMapDef_::iterator it=listaRegioni_.begin(); it!=listaRegioni_.end(); ++it)
 {
  regs.push_back(it->first) ;
 }
 
 sort(regs.begin(), regs.end()) ;

 for(int i=0; i<regs.size(); ++i)
 {
  cout << __LINE__ << "]   " << i+1 << "]\t" << regs[i] << endl ;
 }
} 
//=======================================================================
Analisi::vDef Analisi::getTipiDati(void)
{
 vDef tpdat ;
 for(binMapDef_::iterator it=listaTipoDato_.begin(); it!=listaTipoDato_.end(); ++it)
 {
  tpdat.push_back(it->first) ;
 }
 
 sort(tpdat.begin(), tpdat.end()) ;
 return tpdat ;
} 
//=======================================================================
Analisi::vDef Analisi::getRegioni(void)
{
 vDef regs ;
 for(binMapDef_::iterator it=listaRegioni_.begin(); it!=listaRegioni_.end(); ++it)
 {
  regs.push_back(it->first) ;
 }
 
 sort(regs.begin(), regs.end()) ;
 return regs ;
} 
//=======================================================================
void Analisi::listaTipiDati(void)
{
 cout << __LINE__ << "] Lista dei tipi di dati disponibili" << endl ;
 vector<string> regs ;
 for(binMapDef_::iterator it=listaTipoDato_.begin(); it!=listaTipoDato_.end(); ++it)
 {
  if( it->first == "data_"                  ||
      it->first == "stato_"                 ||
      it->first == "codice_regione_"        ||
      it->first == "denominazione_regione_" ||
      it->first == "latitude_"              ||
      it->first == "longitude_"              ) {continue;}
  regs.push_back(it->first) ;
 }
 
 sort(regs.begin(), regs.end()) ;

 for(int i=0; i<regs.size(); ++i)
 {
  cout << __LINE__ << "]   " << i+1 << "]\t" << regs[i] << endl ;
 }
} 
//=======================================================================
void Analisi::userCode(vDef regione, vDef tipologia)
{
 if( regione.size() > 1 || tipologia.size() > 1 )
 {
  cout << __LINE__ << "] Spiacente, questo codice funziona per un singolo istogramma alla volta\n"
       << __LINE__ << "(ossia una sola regione e una sola tipologia di dati)" << endl ;
 }
 cout << __LINE__ << "] Esempio di utilizzo dei dati:" << endl ;
 for(datMapDef_::iterator it=mapData_.begin(); it!=mapData_.end(); ++it)
 {
  int bin = binMap_[it->first] ;
  for(ssrMapDef_::iterator jt=it->second.begin(); jt!=it->second.end(); jt++)
  {
   bool rejectRegione = true ;
   for(int i=0; i<regione.size(); ++i)
   {
    if(jt->first == regione[i]) rejectRegione = false ;
   }
   if( rejectRegione ) {continue;}
   for(binMapDef_::iterator lt=jt->second.begin(); lt!=jt->second.end(); lt++)
   {
    bool rejectTipologia = true ;
    for(int i=0; i<tipologia.size(); ++i)
    {
     if(lt->first == tipologia[i]) rejectTipologia = false ;
    }
    if( rejectTipologia ) {continue;}
    cout << __LINE__ 
         << "] hisMap_[" 
         << jt->first
         << "]["
         << lt->first
         << "]->SetBinContent("
         << bin
         << ","
         << lt->second
         << ") ;"
         << endl ;
    hisMap_[jt->first][lt->first]->SetBinContent(bin,lt->second) ;
   }
  }
 }
 
// Printout and save
 cout << __LINE__                          << endl ;
 cout << __LINE__ << "] Printout and save" << endl ;
 
 for(hisMapDef_::iterator it=hisMap_.begin(); it!=hisMap_.end(); it++)
 {
  for(ptrMapDef_::iterator jt=it->second.begin(); jt!=it->second.end(); jt++)
  {
   jt->second->Draw("E1") ;
   cout << __LINE__ << "] jt->second->Draw(\"E1\") ; " << endl ;
  }
 }
}
//=======================================================================
Analisi::mapDef Analisi::getData(string regione, string tipologia)
{
 mapDef dataArray ;
 for(datMapDef_::iterator it=mapData_.begin(); it!=mapData_.end(); ++it)
 {
  int bin = binMap_[it->first] ;
  for(ssrMapDef_::iterator jt=it->second.begin(); jt!=it->second.end(); jt++)
  {
   if(jt->first != regione) {continue ;}
   for(binMapDef_::iterator lt=jt->second.begin(); lt!=jt->second.end(); lt++)
   {
    if(lt->first != tipologia) {continue;}
    dataArray[bin] = lt->second ;
   }
  }
 }
 return dataArray ;
}
//=======================================================================
int Analisi::getDataDate(string date, string regione, string tipologia)
{
 int value ;
 for(datMapDef_::iterator it=mapData_.begin(); it!=mapData_.end(); ++it)
 {
  if( it->first != date ) {continue;}
  int bin = binMap_[it->first] ;
  for(ssrMapDef_::iterator jt=it->second.begin(); jt!=it->second.end(); jt++)
  {
   if(jt->first != regione) {continue ;}
   for(binMapDef_::iterator lt=jt->second.begin(); lt!=jt->second.end(); lt++)
   {
    if(lt->first != tipologia) {continue;}
    value = lt->second ;
//     cout << __LINE__ << "] " << __PRETTY_FUNCTION__ << "] " 
//          << it->first<< " - " 
//          << jt->first<< " - " 
//          << lt->first << " (" << tipologia << ") = " 
//          << lt->second << endl;
   }
  }
 }
 return value ;
}
//=======================================================================
void Analisi::makePlot(string regione, string tipologia)
{
 stringstream s; s << regione << "-" << tipologia ;
 TCanvas * canvas = new TCanvas("c","c",1200,1000) ;
 TH1F    * plot   = new TH1F(s.str().c_str(),s.str().c_str(),nBins_,xmi_,xma_) ;
 
 mapDef dataArray = getData(regione, tipologia) ;;
 for(mapDef::iterator it=dataArray.begin(); it!=dataArray.end(); it++)
 {
  plot->SetBinContent(it->first,it->second) ;
 }
 
 plot->Draw("E1") ;
}
//=======================================================================
TH1F * Analisi::getDerivative(string regione, string tipologia)
{
 stringstream s; s << regione << " - " << tipologia << " (al " << this->getLastDate() << ")";
 TH1F    * plot   = new TH1F(s.str().c_str(),s.str().c_str(),nBins_,xmi_,xma_) ;
 plot->GetXaxis()->SetTitle("Giorni a partire dal 2020-02-24" ) ; 
 plot->GetXaxis()->SetTitleColor(4) ; 
 
 mapDef dataArray = getData(regione, tipologia) ;
 int value        = 0 ;
 for(mapDef::iterator it=dataArray.begin(); it!=dataArray.end(); it++)
 {
  value = it->second ;
  if( it!=dataArray.begin())
  {
   value = next(it,1)->second - value ;
   if( value < 0 ) value = 0 ;
  }
  cout << __LINE__ << "] this bin: " << it->second << " next: " << next(it,1)->second << " value: " << value <<  endl ;
  plot->SetBinContent(it->first,value         ) ;
  plot->SetBinError  (it->first,sqrt(value)*5 ) ;
 }
  
 
 return plot;
}
//=======================================================================
TH1F * Analisi::getSinglePlot(string regione, string tipologia)
{
 stringstream s; s << regione << " - " << tipologia << " (al " << this->getLastDate() << ")";
 TH1F    * plot = new TH1F(s.str().c_str(),s.str().c_str(),nBins_,xmi_,xma_) ;
 plot->GetXaxis()->SetTitle("Giorni a partire dal 2020-02-24" ) ; 
 plot->GetXaxis()->SetTitleColor(4) ; 
 
 mapDef dataArray = getData(regione, tipologia) ;
 for(mapDef::iterator it=dataArray.begin(); it!=dataArray.end(); it++)
 {
  plot->SetBinContent(it->first,it->second) ;
 }
 
 return plot;
}
//=======================================================================
TH1F * Analisi::makePlotTotale(string tipologia)
{
 vDef regioni = this->getRegioni() ;
 
 stringstream s; s << "TotaleNazionale-" << tipologia ;
 TCanvas * canvas = new TCanvas("c","c",1200,1000) ;
 TH1F    * plot   = new TH1F(s.str().c_str(),s.str().c_str(),nBins_,xmi_,xma_) ;
 
 double oldContent = 0 ;
 double total      = 0 ;
 mapDef totalArray     ;
  
 for(int i=0; i<regioni.size(); i++)
 {
  mapDef dataArray = getData(regioni[i], tipologia) ;
  for(mapDef::iterator it=dataArray.begin(); it!=dataArray.end(); it++)
  {
   int bin = it->first ; 
   int val = it->second; 
   totalArray[bin] += val ;
  }
 }
 
 for(mapDef::iterator it=totalArray.begin(); it!=totalArray.end(); ++it)
 {
//  cout<< __LINE__ << "] " << 
  plot->SetBinContent(it->first, it->second) ;
 }  
 this->plotAndFit(plot, true) ;
 return plot ;
}
//=======================================================================
string Analisi::getLastDate(void) 
{
 string lastDate = "" ;
 for(datMapDef_::iterator it=mapData_.begin(); it!=mapData_.end(); ++it)
 {
  lastDate = it->first ;
 }
 return lastDate ;
} 
//=======================================================================
Analisi::vDef Analisi::getAllDates(void) 
{
 Analisi::vDef allDates ;
 for(datMapDef_::iterator it=mapData_.begin(); it!=mapData_.end(); ++it)
 {
cout << __LINE__ << "] [" << __PRETTY_FUNCTION__ << "] " << it->first << endl ;
  allDates.push_back(it->first) ;
 }
 return allDates ;
} 
//=======================================================================
string Analisi::formatter(double value, int intPart, int decPart) 
{
 stringstream ss ;
 ss << "%" << intPart << "." << decPart << "f" ;
 char buffer[500];
 sprintf (buffer, ss.str().c_str(), value);
 
 ss.str(""); ss << buffer ;
 return ss.str() ;
}
